@@ -1,19 +1,3 @@
-# This file is part of CairoSVG
-# Copyright © 2010-2018 Kozea
-#
-# This library is free software: you can redistribute it and/or modify it under
-# the terms of the GNU Lesser General Public License as published by the Free
-# Software Foundation, either version 3 of the License, or (at your option) any
-# later version.
-#
-# This library is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
-# details.
-#
-# You should have received a copy of the GNU Lesser General Public License
-# along with CairoSVG.  If not, see <http://www.gnu.org/licenses/>.
-
 """
 Cairo surface creators.
 
@@ -24,7 +8,7 @@ import io
 
 import cairocffi as cairo
 
-from .colors import color
+from .colors import color, negate_color
 from .defs import (
     apply_filter_after_painting, apply_filter_before_painting, clip_path,
     filter_, gradient_or_pattern, linear_gradient, marker, mask, paint_mask,
@@ -32,7 +16,7 @@ from .defs import (
 from .helpers import (
     UNITS, PointError, clip_rect, node_format, normalize, paint,
     preserve_ratio, size, transform)
-from .image import image
+from .image import image, invert_image
 from .parser import Tree
 from .path import draw_markers, path
 from .shapes import circle, ellipse, line, polygon, polyline, rect
@@ -47,6 +31,7 @@ SHAPE_ANTIALIAS = {
 }
 
 TEXT_ANTIALIAS = {
+    'crispEdges': cairo.ANTIALIAS_NONE,
     'optimizeSpeed': cairo.ANTIALIAS_FAST,
     'optimizeLegibility': cairo.ANTIALIAS_GOOD,
     'geometricPrecision': cairo.ANTIALIAS_BEST,
@@ -111,9 +96,10 @@ class Surface(object):
     @classmethod
     def convert(cls, bytestring=None, *, file_obj=None, url=None, dpi=96,
                 parent_width=None, parent_height=None, scale=1, unsafe=False,
-                write_to=None, output_width=None, output_height=None,
-                **kwargs):
-        """Convert a SVG document to the format for this class.
+                background_color=None, negate_colors=False,
+                invert_images=False, write_to=None, output_width=None,
+                output_height=None, **kwargs):
+        """Convert an SVG document to the format for this class.
 
         Specify the input by passing one of these:
 
@@ -127,7 +113,8 @@ class Surface(object):
         :param parent_width: The width of the parent container in pixels.
         :param parent_height: The height of the parent container in pixels.
         :param scale: The ouptut scaling factor.
-        :param unsafe: A boolean allowing XML entities and very large files
+        :param unsafe: A boolean allowing external file access, XML entities
+                       and very large files
                        (WARNING: vulnerable to XXE attacks and various DoS).
 
         Specifiy the output with:
@@ -145,14 +132,17 @@ class Surface(object):
         output = write_to or io.BytesIO()
         instance = cls(
             tree, output, dpi, None, parent_width, parent_height, scale,
-            output_width, output_height)
+            output_width, output_height, background_color,
+            map_rgba=negate_color if negate_colors else None,
+            map_image=invert_image if invert_images else None)
         instance.finish()
         if write_to is None:
             return output.getvalue()
 
     def __init__(self, tree, output, dpi, parent_surface=None,
                  parent_width=None, parent_height=None,
-                 scale=1, output_width=None, output_height=None):
+                 scale=1, output_width=None, output_height=None,
+                 background_color=None, map_rgba=None, map_image=None):
         """Create the surface from a filename or a file-like object.
 
         The rendered content is written to ``output`` which can be a filename,
@@ -176,6 +166,7 @@ class Surface(object):
             self.masks = parent_surface.masks
             self.paths = parent_surface.paths
             self.filters = parent_surface.filters
+            self.images = parent_surface.images
         else:
             self.markers = {}
             self.gradients = {}
@@ -183,6 +174,7 @@ class Surface(object):
             self.masks = {}
             self.paths = {}
             self.filters = {}
+            self.images = {}
         self._old_parent_node = self.parent_node = None
         self.output = output
         self.dpi = dpi
@@ -194,6 +186,16 @@ class Surface(object):
 
         if output_width and output_height:
             width, height = output_width, output_height
+        elif output_width:
+            if width:
+                # Keep the aspect ratio
+                height *= output_width / width
+            width = output_width
+        elif output_height:
+            if height:
+                # Keep the aspect ratio
+                width *= output_height / height
+            height = output_height
         else:
             width *= scale
             height *= scale
@@ -202,6 +204,10 @@ class Surface(object):
         self.cairo, self.width, self.height = self._create_surface(
             width * self.device_units_per_user_units,
             height * self.device_units_per_user_units)
+
+        if 0 in (self.width, self.height):
+            raise ValueError('The SVG size is undefined')
+
         self.context = cairo.Context(self.cairo)
         # We must scale the context as the surface size is using physical units
         self.context.scale(
@@ -209,6 +215,13 @@ class Surface(object):
         # Initial, non-rounded dimensions
         self.set_context_size(width, height, viewbox, tree)
         self.context.move_to(0, 0)
+
+        if background_color:
+            self.context.set_source_rgba(*color(background_color))
+            self.context.paint()
+
+        self.map_rgba = map_rgba
+        self.map_image = map_image
         self.draw(tree)
 
     @property
@@ -260,6 +273,11 @@ class Surface(object):
         """Read the surface content."""
         self.cairo.finish()
 
+    def map_color(self, string, opacity=1):
+        """Parse a color ``string`` and apply ``map_rgba`` function to it."""
+        rgba = color(string, opacity)
+        return self.map_rgba(rgba) if self.map_rgba else rgba
+
     def draw(self, node):
         """Draw ``node`` and its children."""
 
@@ -273,7 +291,7 @@ class Surface(object):
 
         # Do not draw elements with width or height of 0
         if (('width' in node and size(self, node['width']) == 0) or
-           ('height' in node and size(self, node['height']) == 0)):
+                ('height' in node and size(self, node['height']) == 0)):
             return
 
         # Save context and related attributes
@@ -292,7 +310,9 @@ class Surface(object):
         self.context.save()
 
         # Apply transformations
-        transform(self, node.get('transform'))
+        transform(
+            self, node.get('transform'),
+            transform_origin=node.get('transform-origin'))
 
         # Find and prepare opacity, masks and filters
         mask = parse_url(node.get('mask')).fragment
@@ -366,6 +386,7 @@ class Surface(object):
                 self.stroke_and_fill = False
                 self.draw(path)
                 self.stroke_and_fill = True
+                path.tag = 'clipPath'
                 self.context.restore()
                 # TODO: fill rules are not handled by cairo for clips
                 # if node.get('clip-rule') == 'evenodd':
@@ -414,10 +435,11 @@ class Surface(object):
             # Fill
             self.context.save()
             paint_source, paint_color = paint(node.get('fill', 'black'))
-            if not gradient_or_pattern(self, node, paint_source):
-                if node.get('fill-rule') == 'evenodd':
-                    self.context.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
-                self.context.set_source_rgba(*color(paint_color, fill_opacity))
+            if node.get('fill-rule') == 'evenodd':
+                self.context.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
+            if not gradient_or_pattern(self, node, paint_source, fill_opacity):
+                self.context.set_source_rgba(
+                    *self.map_color(paint_color, fill_opacity))
             if TAGS[node.tag] == text:
                 self.cursor_position = save_cursor[0]
                 self.cursor_d_position = save_cursor[1]
@@ -432,9 +454,10 @@ class Surface(object):
             self.context.set_line_width(
                 size(self, node.get('stroke-width', '1')))
             paint_source, paint_color = paint(node.get('stroke'))
-            if not gradient_or_pattern(self, node, paint_source):
+            if not gradient_or_pattern(
+                    self, node, paint_source, stroke_opacity):
                 self.context.set_source_rgba(
-                    *color(paint_color, stroke_opacity))
+                    *self.map_color(paint_color, stroke_opacity))
             self.context.stroke()
             self.context.restore()
         elif not visible:
@@ -482,6 +505,16 @@ class PSSurface(Surface):
     surface_class = cairo.PSSurface
 
 
+class EPSSurface(Surface):
+    """A surface that writes in Encapsulated PostScript format."""
+
+    def _create_surface(self, width, height):
+        """Create and return ``(cairo_surface, width, height)``."""
+        cairo_surface = cairo.PSSurface(self.output, width, height)
+        cairo_surface.set_eps(True)
+        return cairo_surface, width, height
+
+
 class PNGSurface(Surface):
     """A surface that writes in PNG format."""
     device_units_per_user_units = 1
